\ 3iShift.fs -- Shift Register application code based on 3-Instruction Forth for AVR Atmel 328 / Arduino Nano
\ Assad Ebrahim, assadebrahim2000@gmail.com, http://www.mathscitech.org/articles
\ Started: 2021-04-10---Sat---20:27 --- completed 2021-04-10---Sat---20:47   --- 20 minutes to verify the behaviour

\ SHIFT OUT REGISTER (POSI) -- 74595 IC DIP16 -- shift out from controller to chip, parallel out from chip to load pins
\ Chip has two 8-bit registers in memory, a shift register that is filled by pulsing the clock with data bits, which are latched into a storage register.  Output enable transfers the signals from storage register to load.
\ Useful for driving e.g. arrays of LEDs.
\ The output pins Q_A (1) to Q_H (8) can be thought of as an 8-bit FIFO queue with Q_A as entry and Q_H as exit, i.e. after 8 clock cycles, the first bit presented on SER INPUT line (Data In) will reside in Q_H.
\ The programmer needs to decide if Q_A will be interpreted as MSB or LSB of the 8-bit data.
\   This is decided in how data is fed in by controller, and how the Q pins are wired to the load.
\      e.g. using rshift to extract the LSB of a byte and feed into SER INPUT makes Q_A MSB and Q_H LSB.
\      With cascaded shift out registers, Q_A of the first SR is MSB and Q_H of last SR is LSB.
\ Both clock and latch are triggered on the rising edge of a pulse
\ Note - arduino has shiftOut() as a built-in function, but this uses software (bit-banging).  SPI pins would give a hardware behaviour.
\ Note - arduino has millis() for fine grained delay

\ WIRING CONNECTIONS (chip -> Nano, color):
\ see screenshot from chip datasheet
\ IC10 -> nanoA1/15 - CLEAR  grey  ( OPTIONAL --- equivalent to send 0 byte and clock it through )
\ IC11 -> nano13 - CLK  orange
\ IC12 -> nanoA0/14 - LATCH  yellow
\ IC13 -> nano10 - OUTPUT ENABLE  white   ( OPTIONAL --- could just hold output pins always tied to storage reg )
\ IC14 -> nano11 - DATA green
\ IC9  -> cascade --- tie this to IC14 of next chip (DATA) to cascade
\ IC16 -> power
\ IC8  -> ground

\ FIRMWARE - Arduino Nano running 3iForth.avr

\ USAGE:
\ include 3iForth.fs
\ 1 open-com
\ include 3iShift.fs

\ init-shiftout
\ clr

\ V1 = BIT-BANGING IMPLEMENTATION IN SERIAL (SLOW due to serial latency between Forth and Nano).
\ V2 = (TODO) send serial from Forth to Nano and do the bit banging at the Nano level.

\ fundamental functionality
\ : pin-enable 10 ;  \ CS  enables output from storage register to loads
\ : pin-latch  14 ;  
\ : pin-clk    13 ;  \ SCK
\ : pin-data   11 ;  \ COPI
\ : pin-clear  15 ;  \ NOT NEEDED since   0 set-byte   achieves the same (but slower) -- now locked to high (vcc)

16 constant pin-data   16  \ COPI
17 constant pin-clk    17  \ SCK
18 constant pin-latch  18  
20 constant pin-enable 20  \ CS  enables output from storage register to loads
: register-pins  
    pin-data   20 setreg \ DATA
    pin-clk    21 setreg \ CLK
    pin-latch  22 setreg \ LAT
    pin-enable 23 setreg \ CS (out enable)
;

: pulse-hi ( pin -- ) dup hi 1 ms lo ;   \ rising edge pulse (for clk and latch)
: pulse-lo ( pin -- ) dup lo 1 ms hi ;   \ falling edge pulse (for clear)

: enable-output  pin-enable lo ;
: disable-output pin-enable hi ;
: latch       pin-latch pulse-hi ;
\ : clr         pin-clear pulse-lo latch ;  \ wipes shift register to 0.  latch then wipes storage register.
: clk         pin-clk pulse-hi ;

: init-shiftout  \ sets output pins at defaults for 74595 shift out register IC

\    pin-clear out     pin-clear hi    \ ready for pulse-lo
    register-pins
    pin-data out      pin-data lo     \ bit = 0
    pin-clk out       pin-clk lo      \ ready for pulse-hi
    pin-latch out     pin-latch lo    \ ready for pulse-hi
    pin-enable out    enable-output
;

\ manually bit-banging one clock pulse at a time.  Default pin-data state is low.  Latch to see.
: sh-hi  pin-data hi clk pin-data lo ;
: sh-lo  clk ;

: set-bit  ( bit -- ) pin-data pinwrite clk ;  \ sets data bit.  Issue is that this remembers state.
: [set-byte] ( byte --- ) 8 0 do dup 1 and set-bit    1 rshift loop  drop ;
: set-byte ( byte --- ) [set-byte] latch ;

: [set-id] ( id --- ) 1 swap 1 - lshift [set-byte] ;  \ count id from top of output queue (Q_H)
: set-id ( id --- ) [set-id] latch ;  \ count id from top of output queue (Q_H)


: set-id16 ( id -- ) disable-output 
                      dup 8 > if 0 [set-byte] 8 - set-id 
                              else [set-id] 0 set-byte 
                      then enable-output ;
: set-id32 ( id -- ) 
dup 8 <= if [set-id] 0 [set-byte] 0 [set-byte] 0 [set-byte] 
         else dup 16 <= if 0 [set-byte] 8 - [set-id] 0 [set-byte] 0 [set-byte]
                        else dup 24 <= if 0 [set-byte] 0 [set-byte] 16 - [set-id] 0 [set-byte]
                                       else 0 [set-byte] 0 [set-byte] 0 [set-byte] 24 -  [set-id]
                                       then
                        then
         then
         latch ;

: clr 0 set-byte ;

\ BYTE at a time shift-out using C-code on Nano
: shout-byte ( byte -- )   \ send byte over serial and use shift-out register
     xstore CALL_SHIFTOUTBYTE xcall ;
     
     
\ 2021-05-09---Sun---20:24  Adding Piano32 lights 4 cascaded shift out registers
: clear32 disable-output 4 0 do 0 shout-byte loop enable-output ;  \ add/fix disable output-enable output
: 32-on 4 0 do 255 shout-byte loop ; 
: c-shape $89 shout-byte 0 set-bit ;  \ c + e + g in binary, LEDs 1 5 8 counting from left, or 1000 1001b = $89 = 137
: chf clk latch ; \ chord forward
variable ch-pos \ chord position

: chb  ; \ would need to have stored the starting position for the chord to move it back

\ 8-bit TESTS
: test-lights 256 0 do i set-byte loop ; \ show LEDs in binary
: l2r 8 1 do 8 i - set-id 1 ms loop ; 
: r2l 8 1 do i 1 + set-id 1 ms loop ; 
: kit begin l2r r2l key? until ;  \ knight rider style left/right.  Slow because of serial latency between Forth and Nano.

: test-lights-shout
    256 0 do i shout-byte 15 ms loop 
    256 0 do 255 i - shout-byte loop ;

\ 32-bit TESTS
\ $2  shout-byte
\ $6  shout-byte
\ $E  shout-byte
\ $1E shout-byte

: shout32 ( byte-msb byte-3 byte-2 byte-lsb -- )   \ send 4 bytes over serial and use shift-out register
     [newmsg] [store4] tx  \ x_s
     CALL_SHIFTOUT32 xcall ;
: code32  $1E $E $6 $2 ; 
: test32  code32 shout32 ;



\ APPLICATIONS


\ 8-bit ONE-PIP KIT
: shout-id ( id -- )  1 swap 1 - lshift shout-byte ;
: l2r 8 1 do 8 i - shout-id 50 ms loop ; 
: r2l 8 1 do i 1 + shout-id 50 ms loop ; 
: kit begin l2r 160 ms r2l 160 ms key? until ;  \ knight rider style left/right.  Can go super fast.
\ Exercise - could model the individual timing like pendulum, fastest in the middle then slowing at the ends
\ Exercise - send a two or 3 LED pip pulse there and back


\ TWO-PIP KIT
\ Knight rider style left/right. 
\ MUSIC: https://www.youtube.com/watch?v=x8bqWbys3GQ
\ Options: 1pip or 2pips
\ Timing: 
\   0 ms in l2r/r2l loops creates flashing of the outer 6 leds - imperceptible back and forth
\   10 ms  super fast - barely able to count laps
\   30 ms  fast - can now just about count lengths ** good if its only visual
\   70 ms  ** good if accompanied with audible whoosh-whoosh sound to go in time
\ Pause aesthetic:
\   kit1 is ricochet mode, no pause between sweeps
\   kit2 is sticky mode, slight pause between sweeps, light the knight rider car
30 constant wait
5 constant wavelet  \ binary wavelet that will be swept across, 3 is 2-pip, 5 is horned, 1 is 1-pip
: shout-id ( id -- )  wavelet swap 1 - lshift shout-byte ;  \ 3 in binary is the two-pip wavelet that we will sweep acrosss
: l2r 8 1 do 8 i - shout-id wait ms loop ; 
: r2l 7 1 do i 1 + shout-id wait ms loop ; 
: kit begin l2r r2l key? until 0 shout-byte ;  
: kit2 begin l2r wait 2 *  ms r2l wait ms key? until 0 shout-byte ;  
\ With pause at either end, one side is higher to comp to r2l loop less.
\ Exercise - could model the individual timing like pendulum, fastest in the middle then slowing at the ends

: 5-kit begin 5 1 do 5 i - shout-id 50 ms loop  90 ms 5 1 do i 1 + shout-id 50 ms loop 90 ms key? until 0 shout-byte ;

\ cascaded shift registers

: l2r16 16 1 do 16 i - set-id16 wait ms loop ; 
: r2l16 15 1 do i 1 + set-id16 wait ms loop ; 
: kit16 begin l2r16 wait 2 *  ms r2l16 wait ms key? until 0 shout-byte 0 shout-byte clear32 ;


\ ************************
: help-out
    cr ." 74595 SHIFT OUT REGISTER"
    cr ."     use RealTerm to open the connection to the serial port"
    cr ." In GForth:"
    cr ."     include 3iForth  1 open-com   include 3iShift.fs"
    cr ." init-shiftout"
    cr ." clr ( or 0 shout-byte )"
    cr ." low-level:   pin-data hi clk latch"
    cr ." mid-level:   X set-bit latch"
    cr ." high-level:  X set-byte   X set-id"
    cr ." target-level (fastest): X shout-byte  X shout-id"
    cr ." patterns:     kit  kit2  5-kit"
;


\ SHIFT IN REGISTER IC 74165

\ Connections:
\ chip1 -> Nano A3/17 - LOAD/LATCH (yellow)
\ chip2 -> Nano A4/18 - CLOCK (orange)
\ chip10 -> Nano A5/19 - DATA (green)

17 constant pin-load
18 constant pin-clockIN
19 constant pin-dataIN
4  constant buzzer

: init-shiftin
  pin-load out
  pin-clockIN out
  pin-dataIN in
  buzzer out
;  

\ latch inputs
: load-inputs pin-load lo ;
: hold-inputs pin-load hi ;

\ pin-clockIN lo
\ load-inputs
\ hold-inputs
\ pin-clockIN pulse-hi
\ pin-dataIN pinread .
: shiftin-bit  ( -- bit ) pin-dataIN pinread dup . pin-clockIN pulse-hi  ;
: shiftin-byte 0 cr shiftin-bit or .b cr 7 0 do 1 lshift shiftin-bit or .b cr loop ;

\ Q_H (dip 8) emitted first so MSB, Q_A (dip 1) last so LSB 
: sample load-inputs 1 ms hold-inputs shiftin-byte ." MSB Q_H (dip 8): " .b ." :Q_A LSB (dip 1) = " dup . ;  



\ NOTE: don't forget decoupling capacitors on the IC
\ Also - remember need pull-down resistors on the buttons to ensure inputs aren't floating, and to prevent false readings.)
\ Pull down resistor at 10k on 5V supply will draw 1/2mA, so 8 drawing is 4mA.
\ Could use 100k to draw even less at 0.4mA.
\ 1 floating input is ok, but multiple and spurious readings occur

\ APPLICATION
: beats  begin // 4 pulse-hi  key? until ;  \ regular beats
: beats-sequencer ( shift-in -- ) sample begin dup 8 0 do dup 1 and if 4 pulse-hi then // 1 rshift loop drop key? until ;


\ V1 - 2021-04-10---Sat---20:52 --- Added ability to use POSI shift OUT register 74595
\ V2 - 2021-04-12---Mon---03:20 --- Added capability to use PISO shift IN register 74165
\ V3 - 2021-05-09---Sun---20:53 --- ADDED: words for piano32-lights board with 4 cascaded 74595 shift out registers
\       Successfully finished the Electrocookie based PCB prototype: 
\       Complexity: 32 leds, 32 resistors, 4x DIP-16 ICs, 5x capacitors, 6x pin header, 3-pin slide switch, 32 wire connections
\          74 components
\          275 connections hand-soldered connections
\       Defect: the power on switch doesn't really do anything, since power is run through the 4 control pins if the main power is off, it just leads to odd behaviour and probably stresses the chip.  What we probably need is for the keyboard to be electrically isolated from the microcontroller, i.e. we would want to use optoisolator between data pins on the micro and the board.
\ Next Steps:
\ See - 76489_speech_synth.txt   2021-04-09---Fri---04:36
\ V4 - 2021-05-17---Mon---05:11 -- ADDED: beats-sequencer from 8-bit SHIFT-INN register IC 74165s
